import Ember from 'ember';
import Base from 'ember-simple-auth/authenticators/base';
import config from '../config';
import cookieManager from 'allplan-ember-auth/utils/cookiesmanager';
import jQuery from 'jquery';

import {
  getOwner,
} from '@ember/application';
import {
  computed,
  get,
  set,
} from '@ember/object';
import {
  inject as service,
} from '@ember/service';
import {
  resolve,
  reject,
} from 'rsvp';

const allplanAuth = {
  authUrl: config.api.authorize,
  baseUrl: config.api.baseUrl + config.api.namespace,
  appId: config.api.applicationId,
  crossAuthUrl: config.api.crossAuth,
};

export default Base.extend({
  ajax: service(),
  tokenWatchdogTimer : null,

  authUrl: computed(function () {
    const {
      authUrl,
      baseUrl
    } = allplanAuth;
    return `${baseUrl}${authUrl}`;
  }),

  crossAuthUrl: computed(function () {
    const {
      crossAuthUrl,
      baseUrl
    } = allplanAuth;
    return `${baseUrl}${crossAuthUrl}`;
  }),

  restore(authData) {
    let {
      clientId,
      accessToken,
      rememberMeToken,
      tokenType
    } = authData;

    if (accessToken) {
      return this.tokenAuth(accessToken, tokenType)
        .then(() => {
          this.setupCredentials(accessToken, rememberMeToken, tokenType, clientId);
          return authData;
        })
        .catch((e) => {
          if (rememberMeToken) {
            return this.rememberMeToken(rememberMeToken, clientId);
          }

          throw e;
        });
    }

    if (rememberMeToken) {
      return this.rememberMeToken(rememberMeToken, clientId);
    }

    return resolve(authData);
  },

  authenticate({
    username,
    password,
    rememberMe,
    token,
    crossToken,
    tokenType = 'BimPlus'
  }) {
    const origTokenType = tokenType;
    if (token) {
      return this.tokenAuth(token, tokenType).then((response) => {
        let clientId = response.audience ? response.audience : response.user_id; // client_id is called audience in this API call
        let accessToken = token;
        let rememberMeToken = response.remember_me_token;
        let tokenType = response.token_type ? response.token_type : origTokenType;

        this.setupCredentials(accessToken, rememberMeToken, tokenType, clientId);
        return {
          clientId,
          accessToken,
          tokenType
        };
      });
    }

    if (crossToken) {
      return this.crossTokenAuth(crossToken).then((response) => {
        let clientId = response.client_id;
        let accessToken = response.access_token;
        let tokenType = response.token_type;

        this.setupCredentials(accessToken, null, tokenType, clientId);
        return {
          clientId,
          accessToken,
          tokenType
        };
      });
    }

    if (username && password) {
      return this.usernameAuth(username, password, rememberMe).then((response) => {
        let clientId = response.client_id;
        let accessToken = response.access_token;
        let rememberMeToken = response.remember_me_token;
        let tokenType = response.token_type;

        this.setupCredentials(accessToken, rememberMeToken, tokenType, clientId);

        return {
          clientId,
          accessToken,
          tokenType,
          rememberMeToken
        };
      });
    }

    return reject();
  },

  invalidate() {
    return this.get('ajax')
      .delete(this.get('authUrl'))
      .catch((e) => {
        // do nothing for syntax error
        // server responds with invalid JSON and ajax throws error
        if (!(e instanceof SyntaxError)) {
          throw e;
        }
      })
      .finally(() => {
        localStorage.removeItem('accessToken');
        localStorage.removeItem('clientId');
        cookieManager.deleteCookie('ember-auth-rememberable');
        this.stopTokenWatchdog();
      });
  },

  rememberMeToken(rememberMeToken, clientId) {
    const {
      appId
    } = allplanAuth;
    let rememberMe = true;

    return this.get('ajax').post(this.get('authUrl'), {
      data: {
        remember_me_token: rememberMeToken,
        application_id: appId,
        client_id: clientId,
        remember_me: rememberMe
      }
    }).then((response) => {
      let tokenType = response.token_type;
      let accessToken = response.access_token;
      this.setupCredentials(accessToken, rememberMeToken, tokenType, clientId);

      return {
        clientId,
        rememberMeToken,
        rememberMe,
        tokenType,
        accessToken
      };
    }).catch((e) => {
      cookieManager.deleteCookie('ember-auth-rememberable');
      throw e;
    });
  },

  crossTokenAuth(crossToken) {
    return this.get('ajax').post(this.get('crossAuthUrl'), {
      data: {
        cross_token: crossToken
      }
    });
  },

  tokenAuth(token, tokenType = 'BimPlus') {
    return this.get('ajax').request(this.get('authUrl'), {
      headers: {
        'Authorization': `${tokenType} ${token}`
      }
    });
  },

  usernameAuth(username, password, rememberMe) {
    const {
      appId
    } = allplanAuth;
    return this.get('ajax').post(this.get('authUrl'), {
      data: {
        user_id: username,
        password: password,
        application_id: appId,
        remember_me: rememberMe
      }
    });
  },

  setupCredentials(accessToken, rememberMeToken, tokenType = 'BimPlus', clientId = "") {
    let applicationInstance = getOwner(this);
    let apiAdapter = applicationInstance.lookup('apiAdapter:main');
    let configMain = applicationInstance.lookup('config:main');

    set(configMain, 'api.token', accessToken);
    if (tokenType && apiAdapter && typeof apiAdapter.setAccessTokenAndType === 'function') {
      apiAdapter.setAccessTokenAndType(accessToken, tokenType);
    }
    else {
      apiAdapter.setAccessToken(accessToken);
    }
    jQuery.ajaxSetup({
      headers: {
        'Authorization': `${tokenType} ${accessToken}`
      }
    });

    this.setupCredentialsEmbeddedControls(accessToken, tokenType);

    const disabledCookies = configMain && configMain.settings && configMain.settings.noBimCookies;
    // for compatibility with old auth
    if (rememberMeToken && !Ember.testing && !disabledCookies) {
      cookieManager.createCookie('ember-auth-rememberable', rememberMeToken);
    }

    if (rememberMeToken){
      this.initTokenWatchdog(apiAdapter, rememberMeToken, clientId);
    }

  },

  setupCredentialsEmbeddedControls(accessToken, tokenType) {
    if (Ember.testing) {
      return;
    }
    // this is probably needed for embedded controls
    localStorage.setItem('accessToken', accessToken);
    tokenType && localStorage.setItem('tokenType', tokenType);
  },

  /**--------------------------------------------------------------------------
  * initialize token validity watchdog
  * If access token is about to expire and remember me is active, we request new
  * access token before the old one expires - this prevents 401 errors
  --------------------------------------------------------------------------*/
  initTokenWatchdog(apiAdapter, rememberMeToken, clientId){
    const WATCHDOG_TIME_SPAN = 10000; //ms
    apiAdapter.authorize.checkAccessToken().then((retval) => {
      let tokenValid = retval && retval.expires_in && retval.expires_in > 0;
      if (tokenValid) {
        let renewDelay = (retval.expires_in * 1000) - WATCHDOG_TIME_SPAN;
        renewDelay = renewDelay >= 0 ? renewDelay : 50;

        this.stopTokenWatchdog();

        let newTimer = setTimeout(()=>{
          this.rememberMeToken(rememberMeToken, clientId);
        }, renewDelay)

        set(this, 'tokenWatchdogTimer', newTimer);
      }
    });
  },

  /**--------------------------------------------------------------------------
  * stop token watchdog timer for automatic token renewal
  --------------------------------------------------------------------------*/
  stopTokenWatchdog(){
    let tokenWatchdogTimer = get(this, 'tokenWatchdogTimer');

    if ( tokenWatchdogTimer ) {
      clearTimeout(tokenWatchdogTimer);
    }
  }
});
